from pydantic import BaseModel, Field
from typing import Callable, List, Optional, Any

class AgentParams(BaseModel):
    model: str
    api_key: str
    base_url: str
    system_prompt: str = 'You are a helpful AI assistant.'
    summary_prompt: str = 'You are a summarizer that condenses the conversation into a concise summary.'
    tools: Optional[List[Callable]] = Field(default_factory=list)
    thread_id: str
    user_information: Optional[dict[str, Any]] = Field(default_factory=dict)
    temperature: Optional[float] = 0.7



from typing import Literal
from langgraph.graph import StateGraph, START, END
from langgraph.prebuilt import ToolNode, tools_condition
from nodes.nodes import AgentNode, SummarizerNode
from params.agent_params import AgentParams
from state.state import State


class Graph:
    """
    A class to build and compile a state graph for an agentic AI workflow.
    """

    def __init__(self, params: AgentParams):
        self.params = params
        self.graph_builder = StateGraph(State)


    def setup_graph(self, checkpoint=None):
        """
        Set up the graph with a basic chatbot node.
        """

        def agent_conditions(state: State) -> str:
            # First check if tool needs to be called
            tool_result = tools_condition(state)
            if tool_result != END:
                return tool_result

            # Then check if summarizer should be called
            if len(state["messages"]) > 6:
                return "summarizer"

            # Otherwise, finish
            return END

        self.agent_node = AgentNode(params=self.params)
        self.summarizer_node = SummarizerNode(params=self.params)
        self.graph_builder.add_node("agent", self.agent_node.get_agent)
        self.graph_builder.add_node("summarizer", self.summarizer_node.get_agent)
        self.graph_builder.add_node("tools", ToolNode(self.params.tools))
        self.graph_builder.add_edge(START, "agent")
        self.graph_builder.add_conditional_edges(
            "agent",
            agent_conditions
        )
        self.graph_builder.add_edge("tools", "agent")
        self.graph_builder.add_edge("summarizer", END)
        if not checkpoint:
            return self.graph_builder.compile()
        return self.graph_builder.compile(checkpointer=checkpoint)





from typing import Optional

from langgraph.graph.message import add_messages
from typing_extensions import TypedDict, List, Annotated
from langchain_core.messages import BaseMessage
import sys

class State(TypedDict):
    """Represent the structure of the state used in the graph workflow."""
    messages: Annotated[List[BaseMessage], add_messages]
    summary: str
    user_information: dict


def get_size_in_kb(obj) -> float:
    """Recursively calculate size of Python object in kilobytesss."""
    seen_ids = set()

    def sizeof(o):
        if id(o) in seen_ids:
            return 0
        seen_ids.add(id(o))
        size = sys.getsizeof(o)

        if isinstance(o, dict):
            size += sum(sizeof(k) + sizeof(v) for k, v in o.items())
        elif isinstance(o, (list, tuple, set, frozenset)):
            size += sum(sizeof(i) for i in o)

        return size

    total_size_bytes = sizeof(obj)
    return total_size_bytes / 1024  # in KB


from params.agent_params import AgentParams
from base_model.base_rag_model import BaseRagModel
class ChatAI(BaseRagModel):
    def initiate_chatbot(self, params: AgentParams):
        return super().initiate_chatbot(params)



from langchain_core.messages import HumanMessage, RemoveMessage
from llm.baseLLM import BaseLLM
from params.agent_params import AgentParams
from state.state import State

class BaseAgentNode:
    def __init__(self, params: AgentParams):
        self.llm = BaseLLM(params).get_llm()
        self.params = params

    def get_agent(self, state: State):
        pass


class AgentNode(BaseAgentNode):

    def get_agent(self, state: State) -> dict:
        """
        Retrieve the agent's response based on the current state.

        Args:
            state (State): The current state containing messages.

        Returns:
            dict: A dictionary containing the agent's response.
        """
        # print("--- Agent Node Called ---")
        system_message = self.params.system_prompt
        summary = state.get("summary", "")
        if summary:
            system_message += f"Summary of the conversation so far: {summary}\n"

        user_information = state.get("user_information", {})
        response = self.llm.invoke({"system": system_message, "messages": state["messages"], "user_information": user_information})
        return {"messages": [response]}

class SummarizerNode(BaseAgentNode):

    def get_agent(self, state: State) -> dict:
        """
        Summarize the conversation based on the current state.

        Args:
            state (State): The current state containing messages.

        Returns:
            dict: A dictionary containing the summary of the conversation.
        """
        # print("--- Summarizer Node Called ---")
        system_message = self.params.summary_prompt
        summary = state.get("summary", "")
        if summary:
            summary_message = (
                f"This is summary of the conversation to date: {summary}\n\n"
                "Extend the summary by taking into account the new messages above:"
            )
        else:
            summary_message = "Create a summary of the conversation above:"
        user_information = state.get("user_information", {})
        response = self.llm.invoke({"system": system_message, "messages": state["messages"] + [HumanMessage(content=summary_message)], "user_information": user_information})
        delete_messages = [RemoveMessage(id=m.id) for m in state["messages"][:-2]]
        print("delete_messages:", delete_messages)
        return {"messages": delete_messages, "summary": response.content}



from langchain_core.messages import HumanMessage
from langgraph.checkpoint.memory import MemorySaver

from graph.graph import Graph
from params.agent_params import AgentParams


class BaseRagModel:
    def __init__(self):
        self.params: AgentParams
        self.checkpoint = MemorySaver()
        self.config: dict
        self.graph: Graph

    def initiate_chatbot(self, params: AgentParams):
        self.params = params
        self.config = {
            "configurable": {
                "thread_id": self.params.thread_id,
            }
        }
        self.graph = Graph(params=self.params).setup_graph(checkpoint=self.checkpoint)
        return self

    def run(self, messages):
        human_messages = HumanMessage(content=messages)
        response = self.graph.invoke({"messages": [human_messages], "user_information": self.params.user_information}, config=self.config)
        return response["messages"][-1].content





from setuptools import setup, find_packages

setup(
    name='RAGnificentAI',
    version='1.1',
    packages=find_packages(),
    install_requires=[
        "pydantic",
        "langchain",
        "langchain-core",
        "langchain-community",
        "langchain-groq",
        "langchain-openai",
        "langgraph",
        "python-dotenv",
    ],
    author='K. M. Abul Farhad-Ibn-Alam',
    author_email='abulfarhad.ibnalam@gmail.com',
    description='A simple RAG chatbot supporting multiple LLM providers',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Abul-Farhad/RAGnificentAI',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    python_requires='>=3.9',
    keywords='llm chatbot rag openai groq gemini',
    project_urls={
        # 'Bug Reports': 'https://github.com/Abul-Farhad/simple-rag/issues',
        'Source': 'https://github.com/Abul-Farhad/RAGnificentAI',
    },
    license="Custom",
    license_files=('LICENSE',),
)


import os
from dotenv import load_dotenv
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_groq import ChatGroq
from langchain_openai import ChatOpenAI
from params.agent_params import AgentParams

load_dotenv()

class BaseLLM:
    def __init__(self, params: AgentParams):
        self.params = params
        self.llm = ChatOpenAI(
            model=self.params.model,
            api_key=self.params.api_key,
            base_url=self.params.base_url,
            temperature=self.params.temperature
        )
        self.prompt = ChatPromptTemplate.from_messages(
            [
                ("system", "{system}"),
                ("system", "Here is the user information: {user_information}"),
                MessagesPlaceholder("messages"),
            ]
        )

    def get_llm(self):
        chain = self.prompt | self.llm.bind_tools(self.params.tools)
        return chain